// src/SlugTerm.cc
// This file is part of SlugTerm; see http://chezphil.org/slugterm
// (C) 2005 Philip Endecott
// Modifications for 4D Systems Display Graphics Modules (C) 2008 Sven Killig <sven@killig.de>

// This is version $Name:  $
//   (if there is no version (e.g. V0-1) mentioned in the previous line,
//    this is probably a snapshot from between "official" releases.)

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "SlugTerm.hh"

#include <list>
#include <vector>
#include <string>
#include <iostream>

#include <boost/bind.hpp>
#include <boost/mem_fn.hpp>
#include <boost/multi_array.hpp>
#include <boost/thread/xtime.hpp>


using namespace std;


SlugTerm::SlugTerm(string event_dev_fn,
		   KeyboardLayout& keyboard_layout,
		   string device, int bps, int font,
		   commands_t commands,
		   int screenblank_delay):
  keyboard(event_dev_fn,
	   keyboard_layout,
	   boost::bind(&SlugTerm::process_keys,this,_1)),
  lcd(device,bps,font),
  screenblank_delay_(screenblank_delay),
  current_term(0),
  old_screen(boost::extents[lcd.rows()][lcd.cols()]),
  old_screen_attr(boost::extents[lcd.rows()][lcd.cols()])
{
  for(int r=0; r<lcd.rows(); ++r) {
    for(int c=0; c<lcd.cols(); ++c) {
      old_screen[r][c] = 0x20;    /* a space */
      /*int attr;
      ROTE_ATTR_MOD_FG(attr, 7);*/
      old_screen_attr[r][c] = 0x70;  /* white text, black background */
    }
  }
  for (commands_t::const_iterator i = commands.begin();
       i != commands.end(); ++i) {
    Terminal* term = new Terminal(lcd.rows(),lcd.cols(),
				  *i,
				  boost::bind(&SlugTerm::note_term_changed,this));
    /*for(int r=0; r<lcd.rows(); ++r) {
      term.line_dirty_set(r,true);
    }*/
    terms.push_back(term);
  }
}


SlugTerm::~SlugTerm()
{
  for (terms_t::iterator i = terms.begin();
       i != terms.end(); ++i) {
    delete(*i);
  }
}


void SlugTerm::run(void)
{
  boost::thread lcd_update_thread(boost::bind(&SlugTerm::update_lcd,this));

  typedef list<boost::thread*> terminal_threads_t;
  terminal_threads_t terminal_threads;
  for(terms_t::iterator i = terms.begin();
      i != terms.end(); ++i) {
    boost::thread* t = new boost::thread(boost::bind(&Terminal::run,*i));
    terminal_threads.push_back(t);
  }

  keyboard.run();

  for (terminal_threads_t::iterator i = terminal_threads.begin();
       i != terminal_threads.end(); ++i) {
    delete(*i);
  }
}


unsigned int decode_fkey(string key)
{
  if (key=="\033[[A") return 1;
  else if (key=="\033[[B") return 2;
  else if (key=="\033[[C") return 3;
  else if (key=="\033[[D") return 4;
  else if (key=="\033[[E") return 5;
  else if (key=="\033[17~") return 6;
  else if (key=="\033[18~") return 7;
  else if (key=="\033[19~") return 8;
  else if (key=="\033[20~") return 9;
  else if (key=="\033[21~") return 10;
  else if (key=="\033[23~") return 11;
  else if (key=="\033[24~") return 12;
  else return 0;
}


void SlugTerm::process_keys(string keys)
{
  unsigned int fkey = decode_fkey(keys);
  /*if (fkey && fkey<=terms.size()) {
    current_term = fkey-1;
    note_term_changed();
    return;
  }*/

  terms[current_term]->write(keys);
}


void SlugTerm::note_term_changed(void)
{
  boost::mutex::scoped_lock l(pending_changes_mutex);
  pending_changes=true;
  pending_changes_condition.notify_all();
}


void SlugTerm::update_lcd(void)
{
  while(1) {
    {
      boost::mutex::scoped_lock l(pending_changes_mutex);
      if (!pending_changes) {
	struct boost::xtime timeout;
	xtime_get(&timeout, boost::TIME_UTC);
	timeout.sec+=screenblank_delay_;
	if (!pending_changes_condition.timed_wait(l, timeout)) {
	  lcd.backlight(false);
	  lcd.set_mode(false,false,false);
	  continue;
	}
      }
      pending_changes=false;
    }

    Terminal& term = *(terms[current_term]);

    lcd.backlight(true);
    lcd.set_mode(true,false,false);
    unsigned char ch;
    string s;
    bool changed;
    unsigned char attr;
    int cc,cc2;
    for (int r=0; r<lcd.rows(); r++) {
      //s="";
      changed=false;
      cc=0;
      cc2=lcd.cols()-1;
      for (int c=0; c<lcd.cols(); c++) {
	ch   = term.cell(r,c).ch;
	attr = term.cell(r,c).attr;
	// since last send: row, col, s, attr
	// break conditions: change in char/attr WARNING!, 255
	// additional send after loops
	if (old_screen[r][c] != ch || old_screen_attr[r][c] != attr) {
//fprintf(stderr,"changed: %i/%i: %i->%i \n", r, c, old_screen_attr[r][c], attr);
	  //lcd.write_at(r,c,ch,attr);
	    cc=c;
	    changed=true;
	    break;
	}
        //s+=ch;
      }
      for (int c=lcd.cols()-1; c>=cc; c--) {
	ch   = term.cell(r,c).ch;
	attr = term.cell(r,c).attr;
	if (old_screen[r][c] != ch || old_screen_attr[r][c] != attr) {
	    cc2=c;
	    changed=true;
	    break;
	}
      }
      if(changed) /*if(term.line_dirty(r))*/ {
        for (int c=cc; c<=cc2; c++) {
          ch   = term.cell(r,c).ch;
          attr = term.cell(r,c).attr;
          old_screen[r][c]      = ch;
          old_screen_attr[r][c] = attr;
        }
        int c_old=cc;//cc;
        s="";
        unsigned char attr_old = term.cell(r,c_old).attr;
        for (int c2=cc; c2<=cc2; c2++) {
          ch = term.cell(r,c2).ch;
          attr = term.cell(r,c2).attr;
          if(attr_old!=attr/* || c2==lcd.cols()-1*/) {
            lcd.write_at(r,c_old,s,attr_old);
            c_old=c2;
            s="";
            attr_old=attr;
          }
          s+=ch;
        }
        lcd.write_at(r,c_old,s,attr_old);
        //lcd.write_at(r,0,s);
        //term.line_dirty_set(r,false);
      }
    }
    if (term.cur_row()>=0 && term.cur_row()<lcd.rows()
	&& term.cur_col()>=0 && term.cur_col()<lcd.cols()) {
      lcd.set_pos(term.cur_row(), term.cur_col());
      lcd.set_mode(true,true,false);
    }

  }
}
